
{*** Dialog unit for POVSHELL, last modified: 15.Oct.1994, AP ***}

unit povdialg;

interface

  uses Crt, Dos, Objects, Drivers, Memory, Views, Menus, Dialogs,
    StdDlg1, MsgBox1, App, Buffers, PovEdit, PovGlob, PovSHelp;

  type
    TRealInputLine = object(TInputLine)
      LowerBound: real;
      UpperBound: real;
      function Valid(Command: word): boolean; virtual;
      constructor Init(var Bounds: TRect; AMaxLen: Integer;
                       Xmin, Xmax: real);
    end;
    PRealInputLine = ^TRealInputLine;
    TIntegerInputLine = object(TInputLine)
      LowerBound: longint;
      UpperBound: longint;
      function Valid(Command: word): boolean; virtual;
      constructor Init(var Bounds: TRect; AMaxLen: Integer;
                       Xmin, Xmax: longint);
    end;
    PIntegerInputLine = ^TIntegerInputLine;
    TPathInputLine = object(TInputLine)
      function Valid(Command: word): boolean; virtual;
    end;
    PPathInputLine = ^TPathInputLine;
    TFileNameInputLine = object(TInputLine)
      function Valid(Command: word): boolean; virtual;
    end;
    PFileNameInputLine = ^TFileNameInputLine;
    TProgInputLine = object(TFileNameInputLine)
      function Valid(Command: word): boolean; virtual;
    end;
    PProgInputLine = ^TProgInputLine;
    TLibPathInputLine = object(TInputLine)
      function Valid(Command: word): boolean; virtual;
    end;
    PLibPathInputLine = ^TLibPathInputLine;
{    TOutNameInputLine = object(TInputLine)
      function Valid(Command: word): boolean; virtual;
    end;
    POutNameInputLine = ^TOutNameInputLine;}
    TMiscDialog = object(TDialog)
      function Valid(Command: word): boolean; virtual;
    end;
    PMiscDialog = ^TMiscDialog;

  procedure About;

  procedure SetOutputOptions(shift: integer);
  procedure SetDisplayOptions(shift: integer);
  procedure SetTraceOptions(shift: integer);
  procedure SetLibs;
  procedure SetMiscOptions(shift: integer; sel: integer);
  procedure RetrieveOptions;
  procedure SaveOptions;
  procedure DoRetrieveOptions(FileName: PathStr);
  procedure DoSaveOptions(FileName: PathStr);
  procedure SaveDefOptions;
  procedure SetShellDirs;
  procedure SetEditor;
  procedure SetupRun;
  procedure DefineApps;

  procedure ResetOutputOptions;
  procedure ResetDisplayOptions;
  procedure ResetTraceOptions;
  procedure ResetMiscOptions;
  procedure ResetLibs;
  procedure ResetRunOptions;

  procedure ConvertEditorOptions;

  procedure SplitProgram(S: string; var P: PathStr; var O: ComStr);

  function ExtractValue(var S: string; a: integer): string;

implementation

  function TRealInputLine. Valid(Command: word): boolean;
  var
    code: integer;
    x: real;
    l, u: string[20];
    R: TRect;
  begin
    val(Data^,x,code);
    if (code <> 0) or (x < LowerBound) or (x > UpperBound) then begin
      Select; SelectAll(True) ; Valid:=false;
      Str(LowerBound:DataSize-1:2,l);Str(UpperBound:DataSize-1:2,u);
      beep; R.Assign(51,12,77,21);
      if MessageBoxRect(R,'Real Value of range'+#13+
                    l+' to '+u+#13+
                    'required!',nil,mfError+mfOKButton, hcOKButton) = cmOK then;
    end else begin
      Str(x:DataSize-1:2,Data^);
      Valid := TInputLine.Valid(Command);
    end;
  end;

  constructor TRealInputLine.Init(var Bounds: TRect; AMaxLen: Integer;
                   Xmin, Xmax: real);
  begin
    TInputLine.Init(Bounds, AMaxLen);
    LowerBound := Xmin;
    UpperBound := Xmax;
  end;

  function TIntegerInputLine. Valid(Command: word): boolean;
  var
    code: integer;
    x: real; n: longint;
    l, u: string[20];
    R: TRect;
  begin
    val(Data^,x,code); n := Trunc(x);
    if (code <> 0) or (n < LowerBound) or (n > UpperBound) then begin
      Select; SelectAll(True) ; Valid:=false;
      Str(LowerBound,l);Str(UpperBound,u);
      beep; R.Assign(48,12,77,21);
      if MessageBoxRect(R,'Integer Value of range'+#13+
                    l+' to '+u+#13+
                    'required!',nil,mfError+mfOKButton, hcOKButton) = cmOK then;
    end else begin
      Str(n:DataSize-1,Data^);
      Valid := TInputLine.Valid(Command);
    end;
  end;

  constructor TIntegerInputLine.Init(var Bounds: TRect; AMaxLen: Integer;
                   Xmin, Xmax: longint);
  begin
    TInputLine.Init(Bounds, AMaxLen);
    LowerBound := Xmin;
    UpperBound := Xmax;
  end;

  function TPathInputLine. Valid(Command: word): boolean;
  var test: ^PathStr absolute Data;
      R: TRect;
  begin
    if not PathValid(test^) then begin
      Select; SelectAll(True); beep;
      R.Assign(45,15,75,22); Valid := false;
      if MessageBoxRect(R,'This path doesn''t exist!',nil,mfError+mfOKButton, hcOKButton) = cmOK then;
    end else begin
      if test^[length(test^)]='\' then Delete(test^,length(test^),1);
      Valid := TInputLine.Valid(Command);
    end
  end;

  function TFileNameInputLine. Valid(Command: word): boolean;
  var
    R: TRect;
    S: SearchRec;
  begin
    FindFirst(FExpand(Data^), ReadOnly+Archive, S);
    if DosError<>0 then begin
      Select; SelectAll(True); beep;
      R.Assign(49,14,75,22); Valid := false;
      if MessageBoxRect(R,'Invalid filename or'+#13+
          'file doesn''t exist!',nil,mfError+mfOKButton, hcOKButton) = cmOK then;
    end else Valid := TInputLine.Valid(Command);
  end;

  function TProgInputLine. Valid(Command: word): boolean;
  var
    R: TRect;
    S: SearchRec;
    Prog: PathStr;
    i,j: integer;
    D: DirStr; N: NameStr; E: ExtStr;
  begin
    Prog := Trim(Data^);
    if Prog='' then Valid := true
    else begin
      i:=1; while (length(Prog)>i) and (Prog[i+1]<>' ') do inc(i);
      Prog :=copy(Prog,1,i);
      FSplit(Prog,D,N,E); for j:=1 to length(E) do E[j]:=upcase(E[j]);
      if pos('#'+E+'#','#.BAT#.COM#.EXE#')=0 then begin
        Select;
        CurPos:=0; FirstPos:=0; SelStart:=i-4; SelEnd:=i; DrawView; beep;
        R.Assign(49,14,77,22); Valid := false;
        if MessageBoxRect(R,'Invalid extension (File'+#13+
            'not executable)',nil,mfError+mfOKButton, hcOKButton) = cmOK then;
      end else begin
        Prog := FSearch(Prog, GetEnv('PATH'));
        if Prog = '' then begin
          Select;
          CurPos:=0; FirstPos:=0; SelStart:=0; SelEnd:=i; DrawView; beep;
          R.Assign(49,14,75,22); Valid := false;
          MessageBoxRect(R,'Program not found in PATH!',
              nil,mfError+mfOKButton, hcOKButton);
        end else Valid := TInputLine.Valid(Command);
      end;
    end;
  end;

  function TLibPathInputLine. Valid(Command: word): boolean;
  var p,c,i: byte;
      R: TRect;
      S: String[255]; test: PathStr;
      ok: boolean;
  begin
    c := 0; S := Data^; p := 1;
    while (S<>'') and (p>0) and (c<10) do begin
      inc(c); p := pos(';',S);
      if p=0 then LibPath[c] := S else begin
        LibPath[c] := Copy(S,1,p-1); Delete(S,1,p);
      end;
    end;
    ok := true; LibNumber := c;
    for i:=1 to c do ok := ok and PathValid(LibPath[i]);
    if not ok then begin
      Select; SelectAll(True); beep; NoSound;
      R.Assign(42,15,75,22); Valid := false;
      if MessageBoxRect(R,'One or more of the specified'+#13+
                          'paths do not exist!',
                          nil,mfError+mfOKButton, hcOKButton) = cmOK then;
    end else Valid := TInputLine.Valid(Command);
  end;

{  function TOutNameInputLine. Valid(Command: word): boolean;
  const
    IllegalChars = ';,=+<>|"[] \.*? ';
  var
    R: TRect;
  begin
    if Contains(Data^,IllegalChars) or (Data^='') then begin
      Select; SelectAll(True); Sound(1000); Delay(300); NoSound;
      R.Assign(42,13,75,22); Valid := false;
      if MessageBoxRect(R,'Illegal filename! Omit path'+#13+
                          'and extension, check use of'+#13+
                          'special characters!',nil,mfError+mfOKButton,hcOKButton) = cmOK then;
    end else Valid := TInputLine.Valid(Command);
  end;
}
  function TMiscDialog. Valid(Command: word): boolean;
  var MO: MiscOptionsRec;
      R: TRect;
  begin
    if TDialog.Valid(Command) then begin
      GetData(MO);
      if ((MO.StartCol>MO.EndCol) or (MO.StartRow>MO.EndRow))
      and (Command<>cmCancel) then begin
        Valid := false; beep; R.Assign(38,14,75,22);
        if MessageBoxRect(R,' Illegal combination of window'+#13+
                            'coordinates! Adjust the values!',
                            nil,mfError+mfOKButton,hcOKButton) = cmOK then;
      end else Valid := true;
    end else Valid := false;
  end;

  procedure SplitProgram(S: string; var P: PathStr; var O: ComStr);
  var i: integer;
  begin
    P := Trim(S);
    i:=1; while (length(P)>i) and (P[i+1]<>' ') do inc(i);
    if length(P)>i then
      O :=copy(P,i+2,length(P)-i-1)
    else O:='';
    P :=copy(P,1,i);
  end;

  procedure ResetOutputOptions;
  begin
    With OutputOptions do begin
      Format := 1;
      Continue := 1;
      BufferSize := '  -1';
    end;
  end;

  procedure ResetDisplayOptions;
  begin
    With DisplayOptions do begin
      Adapter := 0;
      Palette := 0;
      Size := 7;
      CustomWidth := '  100';
      CustomHeight := '  100';
      ImageWidth := 100;
      ImageHeight := 100;
    end;
  end;

  procedure ResetTraceOptions;
  begin
    With TraceOptions do begin
      Quality := 9;                AntiAliasingCheck := 0;
      AntiAliasingAmount := '0.30'; AAJitterCheck := 1;
      AAJitterAmount := '1.00';     RayNumber := 2;
      BoundingSlabsCheck := 1;     BoundingSlabsThreshold := '   25';
      SymbolTableSize := ' 1000';  VersionCompatibility := 1;
    end;
  end;

  procedure ResetMiscOptions;
  begin
    With MiscOptions do begin
      VPXKOptions := 0;
      ClockValue := '    0.00';
      StartRow := '    1'; EndRow := '  100';
      StartCol := '    1'; EndCol := '  100';
      OtherOptions := '';
    end;
  end;

  procedure ResetLibs;
  begin
    LibNumber := 0;
    Libs.LibPaths:='';
    {???}
  end;

  procedure ResetRunOptions;
  begin
    RunOptions.TheRunOptions := 1;
  end;

  procedure About;
  var
    R: TRect;
    D: PDialog;
  begin
    R.Assign(20, 5,66,19);
    D := New(PDialog, Init(R, ' About '));
    R.Assign( 2, 2,43, 3);
    D^. Insert(New(PStaticText, Init(R, '               POV-Shell '+ProgramVersion)));
    R.Assign( 2, 4,43, 5);
    D^. Insert(New(PStaticText, Init(R, '        The Development Environment')));
    R.Assign( 2, 5,43, 6);
    D^. Insert(New(PStaticText, Init(R, '           to the POV-Raytracer')));
    R.Assign( 2, 7,43, 8);
    D^. Insert(New(PStaticText, Init(R, '    By Andreas Peetz (CIS:[100070,2465])')));
    R.Assign( 2, 9,44,10);
    D^. Insert(New(PStaticText, Init(R, 'This program is Freeware.        '+ProgramDate)));
    R.Assign( 2,10,44,11);
    D^. Insert(New(PStaticText, Init(R, '')));
    R.Assign(19,11,27,13);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    D^.HelpCtx := hcHAbout;
    D^.Options := D^.Options or OfCentered;
    if DeskTop^.ExecView(D) = cmOK then;
    Dispose(D,Done);
  end;

  procedure SetOutputOptions(shift: integer);
  var
    R: TRect;
    D: PDialog;
    BufferSizeString: PIntegerInputLine;
    FormatSelect, ContinueSelect: PMyRadioButtons;
    DummyDir: DirStr;
    DummyExt: ExtStr;
    ValCode,i: integer;
  begin
    R.Assign( 26, 4,53,18);
    D := New(PDialog, Init(R, ' Output-options '));
    R.Assign(14,3,24,5);
    ContinueSelect := New(PMyRadioButtons, Init(R,
      NewSItem('yes',
      NewSItem('no',
      nil))
    ));
    D^. Insert(ContinueSelect);
    R.Assign(13, 2,23, 3);
    D^. Insert(New(PLabel, Init(R, 'Con~t~inue:',ContinueSelect)));
    R.Assign(14,8,20,9);
    BufferSizeString := New(PIntegerInputLine, Init(R, 4, -1, 8192));
    D^. Insert(BufferSizeString);
    R.Assign( 1,8,14,9);
    D^. Insert(New(PLabel, Init(R, '~B~uffersize:',BufferSizeString)));
    R.Assign(2,9,25,10);
    D^.Insert(New(PStaticText, Init(R, '(-1=disable, 0=default)')));
    R.Assign(3,11,11,13);
    D^.Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(14,11,24,13);
    D^.Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(2,3,12,7);
    FormatSelect := New(PMyRadioButtons, Init(R,
      NewSItem('none',
      NewSItem('TGA',
      NewSItem('Dump',
      NewSItem('RAW',
      nil))))
    ));
    D^. Insert(FormatSelect);
    R.Assign( 1, 2, 9, 3);
    D^. Insert(New(PLabel, Init(R, '~F~ormat:',FormatSelect)));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcOOutput;
    D^. SetData (OutputOptions);
    for i:=1 to shift do D^.SelectNext(False);
    if DeskTop^.ExecView(D) <> cmCancel then D^. GetData(OutputOptions);
    Dispose(D,Done);
  end;

  procedure SetDisplayOptions(shift: integer);
  var c,i: integer;
    R: TRect;
    D: PDialog;
    AdapterSelect: PMyRadioButtons;
    PaletteSelect: PMyRadioButtons;
    SizeSelect: PMyRadioButtons;
    CustomWidthString: PIntegerInputLine;
    CustomHeightString: PIntegerInputLine;
    oldWidth,oldHeight: word;
  begin
    R.Assign( 1, 1,77,21);
    D := New(PDialog, Init(R, ' Display-options '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcODisplay;
    R.Assign(2,14,20,19);
    PaletteSelect := New(PMyRadioButtons, Init(R,
      NewSItem('3: 332',
      NewSItem('h: HighColor',
      NewSItem('0: HSV',
      NewSItem('G: GrayScale',
      NewSItem('T: TrueColor',
      nil)))))
    ));
    D^. Insert(PaletteSelect);
    R.Assign(1,13,10,14);
    D^. Insert(New(PLabel, Init(R, '~P~alette:',PaletteSelect)));
    R.Assign(22,14,58,19);
    SizeSelect := New(PMyRadioButtons, Init(R,
      NewSItem('  80 x  50',
      NewSItem(' 160 x 100',
      NewSItem(' 320 x 200',
      NewSItem(' 360 x 480',
      NewSItem(' 640 x 400',
      NewSItem(' 640 x 480',
      NewSItem('1024 x 768',
      NewSItem('Custom:',
      nil))))))))
    ));
    D^. Insert(SizeSelect);
    R.Assign(21,13,27,14);
    D^. Insert(New(PLabel, Init(R, '~S~ize:',SizeSelect)));
    R.Assign(50,17,57,18);
    CustomWidthString := New(PIntegerInputLine, Init(R, 5, 1, 32768));
    D^. Insert(CustomWidthString);
    R.Assign(42,17,50,18);
    D^. Insert(New(PLabel, Init(R, '~W~idth:',CustomWidthString)));
    R.Assign(50,18,57,19);
    CustomHeightString := New(PIntegerInputLine, Init(R, 5, 1, 32768));
    D^. Insert(CustomHeightString);
    R.Assign(42,18,50,19);
    D^. Insert(New(PLabel, Init(R, '~H~eight:',CustomHeightString)));
    R.Assign(61,14,71,16);
    D^.Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(61,17,71,19);
    D^.Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(2,2,74,12);
    AdapterSelect := New(PMyRadioButtons, Init(R,
      NewSItem('Do not display!',
      NewSItem('0: Autodetect (S)VGA type',
      NewSItem('1: Standard VGA 320x200',
      NewSItem('2: Standard VGA 360x480',
      NewSItem('3: Tseng 3000 SVGA 640x480',
      NewSItem('4: Tseng 4000 SVGA',
      NewSItem('5: AT&T VDC600 SVGA 640x400',
      NewSItem('6: Oak Techn. SVGA 640x480',
      NewSItem('7: V7 SVGA 640x480',
      NewSItem('8: V7 Vega (Cirrus) 360x480',
      NewSItem('9: Paradise SVGA 640x480',
      NewSItem('A: Ahead Syst. Ver. A SVGA 640x480',
      NewSItem('B: Ahead Syst. Ver. B SVGA 640x480',
      NewSItem('C: Chips & Techn. SVGA 640x480',
      NewSItem('D: ATI SGVA 640x480',
      NewSItem('E: Everex SVGA 640x480',
      NewSItem('F: Trident SVGA 640x480',
      NewSItem('G: VESA Standard SVGA Adapter',
      NewSItem('H: ATI XL display card',
      NewSItem('I: Diamond SpeedSTAR 24X',
      nil))))))))))))))))))
    ))));
    D^. Insert(AdapterSelect);
    R.Assign( 1,1,10,2);
    D^. Insert(New(PLabel, Init(R, '~A~dapter:',AdapterSelect)));
    D^. SetData(DisplayOptions);
    for i:=1 to shift do D^.SelectNext(False);
    OldWidth := ImageWidth;
    OldHeight := ImageHeight;
    if DeskTop^.ExecView(D) <> cmCancel then D^. GetData(DisplayOptions);
    Dispose(D,Done);
    case Displayoptions.Size of
      0: begin ImageWidth := 80; ImageHeight := 50 end;
      1: begin ImageWidth := 160; ImageHeight := 100 end;
      2: begin ImageWidth := 320; ImageHeight := 200 end;
      3: begin ImageWidth := 360; ImageHeight := 480 end;
      4: begin ImageWidth := 640; ImageHeight := 400 end;
      5: begin ImageWidth := 640; ImageHeight := 480 end;
      6: begin ImageWidth := 1024; ImageHeight := 768 end;
      7: begin val(Displayoptions.CustomWidth,ImageWidth,c);
               val(Displayoptions.CustomHeight,ImageHeight,c);
         end;
    end;
    if (ImageWidth<>OldWidth) or (ImageHeight<>OldHeight) then begin
      str(ImageWidth:5,MiscOptions.EndCol); MiscOptions.StartCol:='    1';
      str(ImageHeight:5,MiscOptions.EndRow); MiscOptions.StartRow:='    1';
    end;
  end;

  procedure SetTraceOptions(shift: integer);
  var
    R: TRect;
    D: PDialog;
    QualitySelect: PMyRadioButtons;
    AntiAliasingCheck: PMyCheckBoxes;
    AntiAliasingAmountLine: PRealInputLine;
    AAJitterCheck: PMyCheckBoxes;
    AAJitterAmountLine: PRealInputLine;
    RayNumberSelect: PMyRadioButtons;
    BoundingSlabsCheck: PMyCheckBoxes;
    BoundingSlabsThresholdLine: PIntegerInputLine;
    SymbolTableSizeLine: PIntegerInputLine;
    VersionCompSelect: PMyRadioButtons;
    i: integer;
  begin
    R.Assign( 9, 3,70,19);
    D := New(PDialog, Init(R, ' Trace-options '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcOTrace;
    R.Assign(19, 2,39,3);
    AntiAliasingCheck := New(PMyCheckBoxes, Init(R,
      NewSItem('~A~nti-Aliasing:',
      nil)
    ));
    D^. Insert(AntiAliasingCheck);
    R.Assign(40,2,46,3);
    AntiAliasingAmountLine := New(PRealInputLine, Init(R, 4, 0.0, 3.0));
    D^. Insert(AntiAliasingAmountLine);
    R.Assign(47,2,60,3);
    D^. Insert(New(PStaticText, Init(R, '(0.0 to 3.0)')));
    R.Assign(19, 4,39,5);
    AAJitterCheck := New(PMyCheckBoxes, Init(R,
      NewSItem('AA-~J~itter:',
      nil)
    ));
    D^. Insert(AAJitterCheck);
    R.Assign(40,4,46,5);
    AAJitterAmountLine := New(PRealInputLine, Init(R, 4, 0.0, 3.0));
    D^. Insert(AAJitterAmountLine);
    R.Assign(47,4,60,5);
    D^. Insert(New(PStaticText, Init(R, '(0.0 to 1.0)')));
    R.Assign(19, 7,54,9);
    RayNumberSelect := New(PMyRadioButtons, Init(R,
      NewSItem('1',
      NewSItem('2',
      NewSItem('3',
      NewSItem('4',
      NewSItem('5',
      NewSItem('6',
      NewSItem('7',
      NewSItem('8',
      NewSItem('9',
      nil)))))))))
    ));
    D^. Insert(RayNumberSelect);
    R.Assign(18,6,54,7);
    D^. Insert(New(PLabel, Init(R, '~N~umber of AA-Rays (to be squared):',RayNumberSelect)));
    R.Assign( 2,10,23,11);
    BoundingSlabsCheck := New(PMyCheckBoxes, Init(R,
      NewSItem('~B~ounding Slabs:',
      nil)
    ));
    D^. Insert(BoundingSlabsCheck);
    R.Assign(25,10,32,11);
    BoundingSlabsThresholdLine := New(PIntegerInputLine, Init(R, 5, 0, 32767));
    D^. Insert(BoundingSlabsThresholdLine);
    R.Assign(33,10,47,11);
    D^. Insert(New(PStaticText, Init(R, '(0 to 32767)')));
    R.Assign( 2,13,9,14);
    SymbolTableSizeLine := New(PIntegerInputLine, Init(R, 5, 301, 32767));
    D^. Insert(SymbolTableSizeLine);
    R.Assign( 1,12,22,13);
    D^. Insert(New(PLabel, Init(R, '~S~ymboltable-size:',SymbolTableSizeLine)));
    R.Assign(10,13,16,14);
    D^. Insert(New(PStaticText, Init(R, '(>300)')));
    R.Assign(22,13,42,14);
    VersionCompSelect := New(PMyRadioButtons, Init(R,
      NewSItem('V1.0',
      NewSItem('V2.0',
      nil))
    ));
    D^. Insert(VersionCompSelect);
    R.Assign(21,12,44,13);
    D^. Insert(New(PLabel, Init(R, '~V~ersion Compatibility:',VersionCompSelect)));
    R.Assign(49,11,59,13);
    D^.Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(49,13,59,15);
    D^.Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(2, 2,16,7);
    QualitySelect := New(PMyRadioButtons, Init(R,
      NewSItem('0',
      NewSItem('1',
      NewSItem('2',
      NewSItem('3',
      NewSItem('4',
      NewSItem('5',
      NewSItem('6',
      NewSItem('7',
      NewSItem('8',
      NewSItem('9',
      nil))))))))))
    ));
    D^. Insert(QualitySelect);
    R.Assign( 1,1,10,2);
    D^. Insert(New(PLabel, Init(R, '~Q~uality:',QualitySelect)));
    D^. SetData(TraceOptions);
    for i:=1 to shift do D^.SelectNext(False);
    if DeskTop^.ExecView(D) <> cmCancel then begin
      D^. GetData(TraceOptions);
    end;
    Dispose(D,Done);
  end;

  procedure SetLibs;
  var
    R: TRect;
    D: PDialog;
    LibPathLine: PLibPathInputLine;
  begin
    R.Assign( 9, 8,70,16);
    D := New(PDialog, Init(R, ' Library-paths '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcOLibPaths;
    R.Assign(17, 5,27, 7);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(33, 5,43, 7);
    D^. Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign( 2, 2,59, 3);
    LibPathLine := New(PLibPathInputLine, Init(R, 255));
    D^. Insert(LibPathLine);
    R.Assign( 2, 3,59, 4);
    D^. Insert(New(PStaticText, Init(R, '(Specify up to 10 paths, separated by ;)')));
    D^. SetData(Libs);
    if DeskTop^.ExecView(D) <> cmCancel then D^. GetData(Libs);
    Dispose(D,Done);
  end;

  procedure SetShellDirs;
  var
    R: TRect;
    D: PDialog;
    DefaultPoVFilePathLine: PPathInputLine;
    OutputPathLine: PPathInputLine;
    ExecPathLine: PProgInputLine;
  begin
    R.Assign( 9, 6,70,17);
    D := New(PDialog, Init(R, ' Directories '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcTDirectories;
    R.Assign(25, 4,59, 5);
    OutputPathLine := New(PPathInputLine, Init(R, 79));
    D^. Insert(OutputPathLine);
    R.Assign( 1, 4,24, 5);
    D^. Insert(New(PLabel, Init(R, 'Default ~O~utput-path:',OutputPathLine)));
    R.Assign(25, 6,59, 7);
    ExecPathLine := New(PProgInputLine, Init(R, 79));
    D^. Insert(ExecPathLine);
    R.Assign( 1, 6,24, 7);
    D^. Insert(New(PLabel, Init(R, 'Path to ~E~xecutable:',ExecPathLine)));
    R.Assign(17, 8,27,10);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(33, 8,43,10);
    D^. Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(25, 2,59, 3);
    DefaultPoVFilePathLine := New(PPathInputLine, Init(R, 79));
    D^. Insert(DefaultPoVFilePathLine);
    R.Assign( 1, 2,24, 3);
    D^. Insert(New(PLabel, Init(R, 'Default ~P~OV-file-path:',DefaultPoVFilePathLine)));
    D^. SetData(Directories);
    if DeskTop^.ExecView(D) <> cmCancel then D^. GetData(Directories);
    Dispose(D,Done);
  end;

  procedure ConvertEditorOptions;
  var c: integer; OldAppVideoMode: word;
  begin
    With EditorOptions do begin
      val(TabSizeString,TabSize,c);
      if (BIASOptions and 1)>0 then
			  EditorFlags:=EditorFlags or efBackupFiles
      else
			  EditorFlags:=EditorFlags and not efBackupFiles;
      OverWrite := not ((BIASOptions and 2)>0);
      AutoIndent := (BIASOptions and 4)>0;
    end;
  end;

  procedure SetEditor;
  var
    TabSizeLine: PIntegerInputLine;
    R: TRect;
    D: PDialog;
    c: integer;
    OldAppVideoMode: word;
  begin
    R.Assign(24,6,56,20);
    D := New(PDialog, Init(R, ' Editor settings '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcTEditor;
    R.Assign(12, 8,16, 9);
    TabSizeLine := New(PIntegerInputLine, Init(R, 2, 2, 16));
    D^. Insert(TabSizeLine);
    R.Assign(1,8,11,9);
    D^. Insert(New(PLabel, Init(R, '~T~ab Size:',TabSizeLine)));
    R.Assign(4,10,14,12);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(17,10,27,12);
    D^. Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(2,2,30,7);
    D^.Insert(New(PMyCheckBoxes, Init(R,
      NewSItem('Create ~B~ackup files',
      NewSItem('~I~nsert mode',
      NewSItem('~A~utoindent mode',
      NewSItem('~S~ave Desktop on exit',
      NewSItem('43/50 ~L~ines',
      nil)))))
    )));
    D^. SetData(EditorOptions);
    if DeskTop^.ExecView(D) <> cmCancel then begin
      D^. GetData(EditorOptions);
      ConvertEditorOptions;
      OldAppVideoMode := AppVideoMode;
      if (EditorOptions.BIASOptions and 16)>0 then
        AppVideoMode := smFont8x8 + smCO80
      else
        AppVideoMode := smCO80;
      if (AppVideoMode <> OldAppVideoMode) then
        Message(DeskTop^.Owner,evBroadCast,cmUpdateScreenMode,nil);
      Message(DeskTop^.Current,evBroadCast,cmUpdateOptions,nil);
      Message(DeskTop^.Current,evBroadCast,cmRedraw,nil);
    end;
    Dispose(D,Done);
  end;

  procedure SetupRun;
  var
    R: TRect;
    D: PDialog;
    c: integer;
  begin
    R.Assign(1,1,33,10);
    D := New(PDialog, Init(R, ' Run Options '));
    D^.Options := D^.Options or ofCentered;
    D^.HelpCtx := hcTRun;
    R.Assign(4,6,14,8);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(17,6,27,8);
    D^. Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    R.Assign(2,2,30,5);
    D^.Insert(New(PMyCheckBoxes, Init(R,
      NewSItem('~P~ause after stats',
      NewSItem('~D~isplay exit status',
      NewSItem('~L~og stats to window',
      nil)))
    )));
    D^. SetData(RunOptions);
    if DeskTop^. ExecView(D) <> cmCancel then D^. GetData(RunOptions);
    Dispose(D,Done);
  end;

  procedure SetMiscOptions(shift: integer; sel: integer);
  const
    kbCode: array[0..4] of word = (kbNoKey,kbAltV,kbAltP,kbAltX,kbAltK);
  var
    R: TRect;
    D: PDialog;
    VPXKCheck: PMyCheckBoxes;
    ClockValueLine: PRealInputLine;
    StartRowInputLine, StartColInputLine: PIntegerInputLine;
    EndRowInputLine, EndColInputLine: PIntegerInputLine;
    OtherOptionsLine: PInputLine;
    i: integer;
    Event: TEvent;
  begin
    R.Assign(20,4,60,20);
    D := New(PMiscDialog, Init(R, ' Miscellaneous options '));
    D^.Options := D^.Options or OfCentered;
    D^.HelpCtx := hcOMisc;
    R.Assign(2,2,38,6);
    VPXKCheck := New(PMyCheckBoxes, Init(R,
      NewSItem('~V~erbose output',
      NewSItem('~P~ause after trace',
      NewSItem('Allow early e~x~it',
      NewSItem('Enable cloc~k~ value:',
      nil))))
    ));
    D^. Insert(VPXKCheck);
    R.Assign(27, 5,37, 6);
    ClockValueLine := New(PRealInputLine, Init(R, 8, -9999.99, 99999.99));
    D^. Insert(ClockValueLine);
    R.Assign(13,8,20,9);
    StartRowInputLine := New(PIntegerInputLine, Init(R, 5,
                         1, ImageHeight));
    D^.Insert(StartRowinputLine);
    R.Assign(1,7,16,8);
    D^. Insert(New(PLabel, Init(R, '~I~mage Window:',StartRowInputLine)));
    R.Assign(2,8,12,9);
    D^. Insert(New(PStaticText, Init(R, 'Start Row:')));
    R.Assign(2,9,12,10);
    D^. Insert(New(PStaticText, Init(R, 'End Row:')));
    R.Assign(21,8,26,9);
    D^. Insert(New(PStaticText, Init(R, 'Col:')));
    R.Assign(21,9,26,10);
    D^. Insert(New(PStaticText, Init(R, 'Col:')));
    R.Assign(26,8,33,9);
    StartColInputLine := New(PIntegerInputLine, Init(R, 5,
                         1, ImageWidth));
    D^.Insert(StartColInputLine);
    R.Assign(13,9,20,10);
    EndRowInputLine := New(PIntegerInputLine, Init(R, 5,
                         1, ImageHeight));
    D^.Insert(EndRowInputLine);
    R.Assign(26,9,33,10);
    EndColInputLine := New(PIntegerInputLine, Init(R, 5,
                         1, ImageWidth));
    D^.Insert(EndColinputLine);
    R.Assign(13,11,37,12);
    OtherOptionsLine := New(PInputLine, Init(R,80));
    D^. Insert(OtherOptionsLine);
    R.Assign(2,11,12,12);
    D^. Insert(New(PLabel, Init(R, 'C~u~stom:',OtherOptionsLine)));
    R.Assign(8,13,18,15);
    D^. Insert(New(PButton, Init(R, '~O~K', cmOK, bfDefault)));
    R.Assign(21,13,31,15);
    D^. Insert(New(PButton, Init(R, '~C~ancel', cmCancel, bfNormal)));
    D^. SetData(MiscOptions);
    for i:=1 to Shift do D^.SelectNext(False);
    if Shift=1 then begin
      Event.What:=evKeyDown;
      Event.KeyCode:=kbCode[Sel];
      VPXKCheck^. HandleEvent(Event);
    end;
    if DeskTop^.ExecView(D) <> cmCancel then D^. GetData(MiscOptions);
    Dispose(D,Done);
  end;

  procedure DefineApps;
  var
    R: TRect;
    D: PDialog;
    N: array[1..MaxApplic] of PInputLine;
  begin
    New(PApplic); if PApplic = Nil then exit;
    for i:=1 to MaxApplic do begin
      PApplic^[i].Name := Applic[i]^.Name^;
      PApplic^[i].Line := ApplicCommand[i].Prog^ + ' ' + ApplicCommand[i].Params^;
    end;
    R.Assign(1,1,76,17);
    D := New(PDialog,Init(R,' Define Applications '));
    D^.options := D^.options or OfCentered;
    D^.HelpCtx := hcUDefineApps;
    R.Assign(10,2,21,3);
    D^.Insert(New(PStaticText,Init(R,'Description')));
    R.Assign(44,2,55,3);
    D^.Insert(New(PStaticText,Init(R,'Commandline')));
    for i:=1 to MaxApplic do begin
      R.Assign(6,1+2*i,24,2+2*i);
      N[i] := New(PInputLine,Init(R,16));
      D^.Insert(N[i]);
      R.Assign(26,1+2*i,73,2+2*i);
      D^.Insert(New(PProgInputLine,Init(R,127)));
      R.Assign(1,1+2*i,5,2+2*i);
      D^.Insert(New(PLabel,Init(R,'#~'+Int2Str(i)+'~:',N[i])));
    end;
    R.Assign(21,13,31,15);
    D^.Insert(New(PButton,Init(R,'~O~K',cmOK,bfDefault)));
    R.Assign(43,13,53,15);
    D^.Insert(New(PButton,Init(R,'~C~ancel',cmCancel,bfNormal)));
    D^.SelectNext(False);
    D^.SetData(PApplic^);
    if DeskTop^.ExecView(D) <> cmCancel then begin
      D^. GetData(PApplic^);
      for i:=1 to MaxApplic do
        if ParseApplic(i,PApplic^[i].Name,PApplic^[i].Line) then
        Desktop^.Owner^.EnableCommands([cmApplic-1+i])
        else Desktop^.Owner^.DisableCommands([cmApplic-1+i]);
    end;
    Dispose(D,Done);
    Dispose(PApplic);
  end;


  function ExtractValue(var S: string; a: integer): string;
  var temp: string[128]; i,j: integer;
  begin temp := ''; i:=0; j := a;
    while (S[j]<>' ') and (i<128) and (j<=Length(S)) do begin
      temp := temp + S[j]; S[j] := ' ';inc(j); inc(i);
    end;
    ExtractValue:=temp;
  end;

  procedure DoRetrieveOptions(FileName: PathStr);
  var
    SR: SearchRec;
    f: Text;
    O: string;
    Line: string;
    i,p,q,L: integer;
    C: char;
  function FindOpt(t: char; O: string): integer;
  var temp: integer;
  begin
    temp := pos(' +'+t,O); if temp = 0 then temp := pos(' -'+t,O);
    FindOpt := temp;
  end;
  procedure BlankOut(var s: string; i,c: integer);
  var j: integer;
  begin
    for j:=i to i+c-1 do s[j] := ' ';
  end;
  begin
    ResetOutputOptions;
    ResetDisplayOptions;
    ResetTraceOptions;
    ResetMiscOptions;
    ResetLibs;
    if pos(TempDEF,FileName)=0 then CurDEFFileName := FileName;
    Assign(f,FileName);
    Reset(f); O := ' ';
    While not eof(f) do begin
      readln(f,Line);
      O := O + Line + ' ';
    end;
    Close(f); L := length(O);
    for i:=1 to L do O[i]:=Upcase(O[i]);
    p := pos(' +A',O); if p>0 then begin
      TraceOptions.AntiAliasingCheck := 1;
      if O[p+3]<>' ' then TraceOptions.AntiAliasingAmount := ExtractValue(O,p+3);
      BlankOut(O,p+1,2);
    end;
    p := FindOpt('B',O); if p>0 then begin
      if O[p+3]=' ' then OutputOptions.BufferSize :='   4'
                    else OutputOptions.BufferSize := ExtractValue(O,p+3);
      BlankOut(O,p+1,2);
    end;
    p := pos(' +C',O); if p>0 then begin
      OutputOptions.Continue := 0;
      BlankOut(O,p+1,2);
    end;
    p := pos(' +D',O); if p>0 then begin
      BlankOut(O,p+1,2);
      if O[p+3]<>' ' then begin
        DisplayOptions.Adapter := pos(O[p+3],Adapters);
        BlankOut(O,p+3,1);
        if O[p+4]<>' ' then begin
          DisplayOptions.Palette := pos(O[p+4],Palettes)-1;
          BlankOut(O,p+4,1);
        end else
          DisplayOptions.Palette := pos('3',Palettes)-1;
      end else begin
        DisplayOptions.Adapter := pos('0',Adapters);
        DisplayOptions.Palette := pos('3',Palettes)-1;
      end;
    end;
    p := FindOpt('F',O); if p>0 then begin
      if O[p+1]='-' then OutputOptions.Format:=0
        else OutputOptions.Format:=pos(O[p+3],Formats);
      BlankOut(O,p+1,3);
    end;
    p := FindOpt('H',O); if p>0 then begin
      BlankOut(O,p+1,2);
      MiscOptions.Endrow := ExtractValue(O,p+3);
      Val(Miscoptions.Endrow,ImageHeight,i);
    end;
    p := FindOpt('J',O); if p>0 then begin
      BlankOut(O,p+1,2);
      if O[p+1]='-' then TraceOptions.AAJitterCheck := 0
      else if O[p+3]<>' ' then TraceOptions.AAJitterAmount := ExtractValue(O,p+3);
    end;
    p := FindOpt('K',O); if p>0 then begin
      BlankOut(O,p+1,2);
      MiscOptions.VPXKOptions := MiscOptions.VPXKOptions or 8;
      MiscOptions.ClockValue := ExtractValue(O,p+3);
    end;
    LibNumber := 0;
    Libs.LibPaths := '';
    repeat
      p := FindOpt('L',O); if p>0 then begin
        inc(LibNumber); LibPath[LibNumber] := ExtractValue(O,p+3);
        Delete(O,p,Length(LibPath[LibNumber])+3);
        Libs.LibPaths := Libs.LibPaths + LibPath[LibNumber] + ';';
      end;
    until (p=0) or (LibNumber=10);
    Delete(Libs.LibPaths, Length(Libs.LibPaths),1);
    p := pos(' -MB',O); if p>0 then begin
      TraceOptions.BoundingSlabsCheck := 0;
      BlankOut(O,p+1,3);
    end;
    p := pos(' +MB',O); if p>0 then begin
      BlankOut(O,p+1,3);
      TraceOptions.BoundingSlabsThreshold := ExtractValue(O,p+4);
    end;
    p := pos(' -MS',O); if p=0 then p := pos(' +MS',O); if p>0 then begin
      BlankOut(O,p+1,3);
      TraceOptions.SymbolTableSize := ExtractValue(O,p+4);
    end;
    p := pos(' -MV1.0',O); if p=0 then p := pos(' +MV1.0',O); if p>0 then begin
      BlankOut(O,p+1,6);
      TraceOptions.VersionCompatibility := 0;
    end;
    p := pos(' +P',O); if p > 0 then begin
      BlankOut(O,p+1,2);
      MiscOptions.VPXKOptions := MiscOptions.VPXKOptions or 2;
    end;
    p := FindOpt('Q',O); if p>0 then begin
      q := pos(O[p+3],Qualities);
      if q>0 then TraceOptions.Quality := q-1;
      BlankOut(O,p+1,3);
    end;
    p := FindOpt('R',O); if p>0 then begin
      q := pos(O[p+3],AARays);
      if q>0 then TraceOptions.RayNumber := q-1;
      BlankOut(O,p+1,3);
    end;
    p := pos(' +V',O); if p > 0 then begin
      MiscOptions.VPXKOptions := MiscOptions.VPXKOptions or 1;
      BlankOut(O,p+1,2);
    end;
    p := pos(' +X',O); if p > 0 then begin
      MiscOptions.VPXKOptions := MiscOptions.VPXKOptions or 4;
      BlankOut(O,p+1,2);
    end;
    p := FindOpt('W',O); if p>0 then begin
      BlankOut(O,p+1,2);
      MiscOptions.EndCol := ExtractValue(O,p+3);
      Val(Miscoptions.EndCol,ImageWidth,i);
    end;
    p := pos(' +EC',O); if p=0 then p := pos(' -EC',O); if p>0 then begin
      if O[p+4]<>' ' then MiscOptions.EndCol := ExtractValue(O,p+4);
      BlankOut(O,p+1,3);
    end;
    p := pos(' +ER',O); if p=0 then p := pos(' -ER',O); if p>0 then begin
      if O[p+4]<>' ' then MiscOptions.EndRow := ExtractValue(O,p+4);
      BlankOut(O,p+1,3);
    end;
    p := FindOpt('E',O);
    if (p>0) and (O[p+3]<>'C') and (O[p+3]<>'R') and (O[p+3]<>' ') then begin
      BlankOut(O,p+1,2);
      MiscOptions.EndRow := ExtractValue(O,p+3);
    end;
    p := pos(' +SC',O); if p=0 then p := pos(' -SC',O); if p>0 then begin
      if O[p+4]<>' ' then MiscOptions.StartCol := ExtractValue(O,p+4);
      BlankOut(O,p+1,3);
    end;
    p := pos(' +SR',O); if p=0 then p := pos(' -SR',O); if p>0 then begin
      if O[p+4]<>' ' then MiscOptions.StartRow := ExtractValue(O,p+4);
      BlankOut(O,p+1,3);
    end;
    p := FindOpt('S',O);
    if (p>0) and (O[p+3]<>'C') and (O[p+3]<>'R') and (O[p+3]<>' ') then begin
      BlankOut(O,p+1,2);
      MiscOptions.StartRow := ExtractValue(O,p+3);
    end;
    MiscOptions.OtherOptions := Trim(O);
    with DisplayOptions do begin
      case ImageWidth of
        80: if ImageHeight = 50 then Size := 0;
        160: if ImageHeight = 100 then Size := 1;
        320: if ImageHeight = 200 then Size := 2;
        360: if ImageHeight = 480 then Size := 3;
        640: if ImageHeight = 400 then Size := 4
             else if ImageHeight = 480 then Size := 5;
        1024:if ImageHeight = 768 then Size := 6;
      end;
      if Size=7 then begin
        Str(ImageWidth:5,CustomWidth);
        StR(Imageheight:5,CustomHeight);
      end;
    end;
  end;

  procedure RetrieveOptions;
  var
    Filename: PathStr;
    SR: SearchRec;
    D: PFileDialog;
    CD: DirStr; CN: NameStr; CE: ExtStr;
  begin
    if (CurDEFFileName<>'')  then begin
      FSplit(CurDEFFileName,CD,CN,CE); Dec(CD[0]);
      ChDir(CD);
      Filename := CN+CE;
    end else begin
      FileName:='*.DEF';
      ChDir(ShellPath);
    end;
    D := New(PFileDialog, Init('*.DEF', 'Retrieve options',
    '~N~ame', fdOpenButton, 101));
    D^.HelpCtx := hcORetrieveOptions;
    D^.SetData(FileName);
    if DeskTop^.ExecView(D) <> cmCancel then begin
      D^.GetData(FileName);
      FindFirst(FileName,AnyFile,SR);
      if DosError<>0 then begin
        beep;
        if MessageBox('File '+FileName+' not found!',nil,mfError+mfOKButton,hcOKButton)=cmOK then;
      end else
        DoRetrieveOptions(FileName);
    end;
    Dispose(D,Done);
  end;

  procedure DoSaveOptions(FileName: PathStr);
  var
    f: Text;
    i: integer;
    t1,t2: string[5];
  begin
    if pos(TempDEF,FileName)=0 then CurDEFFileName := FileName;
    assign(f,FileName); rewrite(f);
    if TraceOptions.AntiAliasingCheck = 1 then
      writeln(f,'+A'+Trim(TraceOptions.AntiAliasingAmount));
    if OutputOptions.BufferSize <> '  -1' then
         writeln(f,'+B'+Trim(OutputOptions.Buffersize));
    if OutputOptions.Continue = 0 then writeln(f,'+C');
    if DisplayOptions.Adapter<>0 then
      writeln(f,'+D'+Adapters[DisplayOptions.Adapter]+
                     Palettes[DisplayOptions.Palette+1]);
    if OutputOptions.Format = 0 then writeln(f,'-F')
    else if OutputOptions.Format <> 1 then
    writeln(f,'+F'+Formats[OutputOptions.Format]);
    if (TraceOptions.AntiAliasingCheck = 1) then
      if (TraceOptions.AAJitterCheck = 0)
        then writeln(f,'-J')
        else if TraceOptions.AAJitterAmount<>'1.00' then
        writeln(f,'+J'+Trim(TraceOptions.AAJitterAmount));
    if (MiscOptions.VPXKOptions and 8) > 0 then
      writeln(f,'+K'+Trim(MiscOptions.ClockValue));
    for i:=1 to LibNumber do
      writeln(F,'+L'+LibPath[i]);
    if TraceOptions.BoundingSlabsCheck = 0 then
      writeln(f,'-MB') else if TraceOptions.BoundingSlabsThreshold<>'   25' then
    writeln(f,'+MB'+Trim(TraceOptions.BoundingSlabsThreshold));
    if TraceOptions.SymbolTableSize<>' 1000' then
      writeln(f,'+MS'+Trim(TraceOptions.SymbolTableSize));
    if TraceOptions.VersionCompatibility = 0 then
      writeln(f,'+MV1.0');
    if (MiscOptions.VPXKOptions and 2) > 0 then
      writeln(f,'+P');
    if TraceOptions.Quality <> 9 then
      writeln(f,'+Q'+chr(48+TraceOptions.Quality));
    if (TraceOptions.AntiAliasingCheck = 1) and (TraceOptions.RayNumber<>2) then
      writeln(f,'+R'+chr(49+TraceOptions.RayNumber));
    if (MiscOptions.VPXKOptions and 1) > 0 then
      writeln(f,'+V');
    if (MiscOptions.VPXKOptions and 4) > 0 then
      writeln(f,'+X');
    str(ImageWidth:5,t1); str(ImageHeight:5,t2);
    writeln(f,'+W'+Trim(t1)+' +H'+Trim(t2));
    with MiscOptions do begin
      if StartCol<>'    1' then writeln(f,'+SC'+Trim(MiscOptions.StartCol));
      if Trim(EndCol)<>Trim(t1) then writeln(f,'+EC'+Trim(MiscOptions.EndCol));
      if StartRow<>'    1' then writeln(f,'+SR'+Trim(MiscOptions.StartRow));
      if Trim(EndRow)<>Trim(t2) then writeln(f,'+ER'+Trim(MiscOptions.EndRow));
      if Trim(OtherOptions)<>'' then writeln(f,OtherOptions);
    end;
    close(f);
  end;

  procedure SaveOptions;
  var
    Filename: PathStr;
    SR: SearchRec;
    D: PFileDialog;
    f: Text;
    CD: DirStr; CN: NameStr; CE: ExtStr;
  begin
    if CurDEFFileName<>'' then begin
      FSplit(CurDEFFileName,CD,CN,CE); Dec(CD[0]);
      ChDir(CD);
      Filename := CN+CE;
    end else begin
      FileName:='*.DEF';
      ChDir(ShellPath);
    end;
    D := New(PFileDialog, Init('*.DEF', 'Save options',
    '~N~ame', fdOpenButton, 101));
    D^.SetData(FileName);
    D^.HelpCtx := hcOSaveOptions;
    if DeskTop^.ExecView(D) <> cmCancel then begin
      D^.GetData(FileName);
      FindFirst(FileName,AnyFile,SR);
      if DosError=0 then begin
        beep;
        if MessageBox('File '+Filename+' exists! OK to overwrite?'
        ,nil,mfWarning+mfOKButton+mfCancelButton, hcOKCancel) = cmCancel
        then exit;
      end;
      DoSaveOptions(FileName);
    end;
    Dispose(D,Done);
  end;

  procedure SaveDefOptions;
  begin
    DoSaveOptions(ShellPath+'\DEFAULT.DEF')
  end;

end.
