{*** global data for POVSHELL
     last modified: 18.Oct.1994, AP ***}

unit PovGlob;

interface

uses dos, crt, objects, dialogs, overlay, stddlg1, povshelp, App,
      views, msgbox1, drivers, memory, buffers, menus;

  {$I commands.inc}

  const
    ProgramVersion = '3.0';
    ProgramDate = 'Oct. 1994';

  const
    TempPOV = '$SHELL$.POV';
    LogFile = 'POVRAY.LOG';

  const
    Adapters: string[19] = '0123456789ABCDEFGHI';
    Palettes: string[5] = '3H0GT'; Formats: string[3] = 'TDR';
    Qualities: string[10] = '0123456789'; AARays: string[9] = '123456789';
    TempDEF: string[11] = '$SHELL$.DEF';

  const
    MaxRememberedTopics = 100;

  const
    FormatExt: array[1..3] of ExtStr = ('.TGA','.DIS','.B8');

  const
    MaxApplic = 5;

  type
    String40 = string[40];

  type
    TPLString = String[12];

  type
    OutputOptionsRec = record
      Continue: word;
      BufferSize: String[4];
      Format: word;
    end;

    DisplayOptionsRec = record
      Palette: word;
      Size: word;
      CustomWidth, CustomHeight: String[5];
      Adapter: word;
    end;

    TraceOptionsRec = record
      AntiAliasingCheck: word;
      AntiAliasingAmount: string[4];
      AAJitterCheck: word;
      AAJitterAmount: string[4];
      RayNumber: word;
      BoundingSlabsCheck: word;
      BoundingSlabsThreshold: string[5];
      SymbolTableSize: string[5];
      VersionCompatibility: word;
      Quality: word;
    end;

    LibsRec = record
      LibPaths: string[255];
    end;

    ShellDirsRec = record
      OutputFilePath: PathStr;
      ExecPath: PathStr;
      DefaultPoVFilePath: PathStr;
    end;

    MiscOptionsRec = record
      VPXKOptions: word;
      ClockValue: string[8];
      StartRow, StartCol: string[5];
      EndRow, EndCol: string[5];
      OtherOptions: string[80];
    end;

    EditorOptionsRec = record
      TabSizeString: string[2];
      BIASOptions: word;
    end;

    RunOptionsRec = record
      TheRunOptions: word;
    end;

    AniOptionsRec = record
      ClockFrom: string[11];
      ClockTo: string[11];
      ClockStep: string[11];
    end;

    ApplicCommandRec = record
      Prog,Params: PString;
    end;

    DefApplicRec = array[1..MaxApplic] of record
                                            Name: string[16];
                                            Line: ComStr;
                                          end;
    PDefApplicRec = ^DefApplicRec;

  type

    TMyRadioButtons = object(TRadioButtons)
      function GetHelpCtx: word; virtual;
    end;
    PMyRadioButtons = ^TMyRadioButtons;

    TMyCheckBoxes = object(TCheckBoxes)
      function GetHelpCtx: word; virtual;
    end;
    PMyCheckBoxes = ^TMyCheckBoxes;

  var
    OutputOptions: OutputOptionsRec;
    DisplayOptions: DisplayOptionsRec;
    TraceOptions: TraceOptionsRec;
    Directories: ShellDirsRec;
    Libs: LibsRec;
    MiscOptions: MiscOptionsRec;
    EditorOptions: EditorOptionsRec;
    RunOptions: RunOptionsRec;
    AniOptions: AniOptionsRec;
    LibPath: Array[1..10] of PathStr;
    LibNumber: integer;
    ImageWidth, ImageHeight: word;
    ShellPath: DirStr;
    CurDEFFileName: PathStr;
    CurPovFileDir: DirStr;
    TabSize: byte;
    OverWrite: boolean;
    AutoIndent: boolean;

    TempPath: String[79];
    EdFileName: PathStr;
    ImageFileName: PathStr;
    StartDirectory: DirStr;
    TPLTopic: TPLString;

    _D: DirStr; _N: NameStr; _E: ExtStr;

    UseMono: boolean;
    CFGPalette: string;

    AppVideoMode: word;

    RememberList: array[0..MaxRememberedTopics] of
                     record KR: word;Sel: integer end;
    Remembered: 0..MaxRememberedTopics;

    Applic: array[1..MaxApplic+1] of PMenuItem;
    ApplicCommand: array[1..MaxApplic] of ApplicCommandRec;
    PApplic: PDefApplicRec;

    i: integer;

  procedure beep;
  procedure RememberTopic(NewKeyRef: word;NowSelected: integer);
  function ExpandOptions(inp: ComStr; var out: ComStr): word;
  function CheckOutput(ImageFileName: FNameStr): word;
  function AskForOutputFile(InputFileName: PathStr; var ImageFileName: PathStr): word;
  procedure ExecProgram(Prog: PathStr; Params: ComStr; WaitForKey: boolean);
  function ParseApplic(i: integer; name: string40; line: string): boolean;
  function trim(s: string): string;
  function Int2Str(i: integer): NameStr;
  procedure ReplaceStr(var S: PString; V: String);

implementation

  procedure beep;
  begin
    Sound(1000); Delay(300); NoSound;
  end;

  procedure RememberTopic(NewKeyRef: word; NowSelected: integer);
  begin
    if Remembered <= MaxRememberedTopics then begin
      {beep;}
      if NowSelected <> 0 then
        RememberList[Remembered].Sel:=NowSelected;
      Inc(Remembered);
      RememberList[Remembered].KR:=NewKeyRef;
    end else Remembered := 0;
  end;

  function ExpandOptions(inp: ComStr; var out: ComStr): word;
  var i,j: integer;
      ParStr: string40;
  begin
    Message(DeskTop^.Current, evBroadCast, cmGetFileName, nil);
    ExpandOptions := cmOK;
    out := inp;
    repeat
      i := pos('%i',out);
      if i>0 then out:=copy(out,1,i-1)+
                  EdFileName+copy(out,i+2,length(out)-i-1);
    until i=0;
    repeat
      i := pos('%o',out);
      if i>0 then out:=copy(out,1,i-1)+
                  ImageFileName+copy(out,i+2,length(out)-i-1);
    until i=0;
    j := 1;
    repeat
      i := pos('%?',out);
      if i>0 then begin
        ParStr := '';
        if InputBox('Call Application','~P~arameter '+Int2Str(j)+':', ParStr,40) <> cmCancel then
          out := copy(out,1,i-1) + ParStr + copy(out,i+2,length(out)-i-1)
        else begin
          ExpandOptions := cmCancel;
          i := 0;
        end;
        inc(j);
      end;
    until i=0;
  end;

  function TMyRadioButtons.GetHelpCtx: word;
  begin
    GetHelpCtx := TRadioButtons.GetHelpCtx - Sel;
  end;

  function TMyCheckBoxes.GetHelpCtx: word;
  begin
    GetHelpCtx := TCheckBoxes.GetHelpCtx - Sel;
  end;

  function CheckOutput(ImageFileName: FNameStr): word;
  var SR: SearchRec;
  begin
    CheckOutput := cmOK;
    FindFirst(ImageFileName,AnyFile,SR);
    if (DosError=0) and (OutputOptions.Continue=1) then begin
      beep;
      if MessageBox('Output-file '+ImageFileName+' exists '+
        'and CONTINUE is set to NO! '+
        'OK to overwrite?',nil,mfWarning+mfOKButton+mfCancelButton, hcOKCancel) = cmCancel
      then CheckOutput:=cmCancel;
    end;
    if (DosError<>0) and (OutputOptions.Continue=0) then begin
      beep;
      if MessageBox('CONTINUE is set to YES, but file '+#13+
                     ImageFileName+' doesn''t exist!'+#13+
        'OK to create new file?',nil,mfWarning+mfOKButton+mfCancelButton, hcOKCancel) = cmCancel
      then CheckOutput:=cmCancel
      else OutputOptions.Continue:=1;
    end;
  end;

  function AskForOutputFile(InputFileName: PathStr; var ImageFileName: PathStr): word;
  var
    D: PFileDialog;
  begin
    FSplit(InputFileName,_D,_N,_E);
    if Directories.OutputFilePath='' then ChDir(Copy(_D,1,Length(_D)-1))
    else ChDir(Directories.OutputFilePath);
    ImageFileName := _N + FormatExt[OutputOptions.Format];
    D := New(PFileDialog, Init('*'+FormatExt[OutputOptions.Format], 'Select output-file',
    '~N~ame', fdOpenButton, 102));
    D^.SetData(ImageFileName);
    D^.HelpCtx := hcFSelectBox;
    if DeskTop^. ExecView(D) = cmCancel then begin
      Dispose(D,Done);
      AskForOutputFile:=cmCancel;
      exit
    end else begin
      D^.GetData(ImageFileName);
      Dispose(D,Done);
      {AskForOutputFile := CheckOutput(ImageFileName);}
      AskForOutputFile := cmOK;
    end;
  end;

  procedure ExecProgram(Prog: PathStr; Params: ComStr; WaitForKey: boolean);
  var EParams: PathStr;
      i: integer;
  begin
    Prog := FSearch(Prog,GetEnv('PATH'));
    Prog := FExpand(Prog);
    if ExpandOptions(Params,EParams)=cmCancel then exit;
    FSplit(Prog,_D,_N,_E); for i:=1 to length(_E) do _E[i]:=upcase(_E[i]);
    DoneSysError;
    DoneEvents;
    DoneVideo;
    DoneMemory;
    SetMemTop(Ptr(BufHeapPtr, 0));
    SwapVectors;
    if _E='.BAT' then
      Exec(GetEnv('COMSPEC'), '/C '+ Prog + ' ' + EParams)
    else
      Exec(Prog,EParams);
    while keypressed do if readkey=#0 then;
    if WaitForKey then begin
      Writeln; Write('Press any key to continue...');
      repeat until keypressed; if readkey=#0 then if readkey=#0 then;
    end;
    SwapVectors;
    SetMemTop(Ptr(BufHeapEnd, 0));
    InitMemory;
    InitVideo;
    InitEvents;
    InitSysError;
    Message(DeskTop^.Owner,evBroadCast,cmRedrawApp,nil);
  end;

  function trim(s: string): string;
  var i: integer; temp: string;
  begin
    temp := s;
    while (length(temp)>0) and (temp[1]=' ') do Delete(temp,1,1);
    while (length(temp)>0) and (temp[length(temp)]=' ') do Delete(temp,Length(temp),1);
    Trim := temp;
  end;

  function ParseApplic(i: integer; name: string40; line: string): boolean;
  var
    c: integer;
  begin
    name := Trim(name); line := Trim(line);
    if (name='') or (line='') then begin
      ReplaceStr(ApplicCommand[i].Prog,' ');
      ReplaceStr(ApplicCommand[i].Params,' ');
      ReplaceStr(Applic[i]^.Name,'(not defined)');
      ParseApplic := false;
    end else begin
      ReplaceStr(Applic[i]^.Name,name);
      c := pos(' ',line);
      if c=0 then begin
         ReplaceStr(ApplicCommand[i].Params,' ');
         c := length(line)+1;
      end else
        ReplaceStr(ApplicCommand[i].Params,copy(line,c+1,length(line)-c));
      ReplaceStr(ApplicCommand[i].Prog,copy(line,1,c-1));
      ParseApplic := true;
    end;
  end;

  function Int2Str(i: integer): NameStr;
  var temp: NameStr;
  begin
   Str(i,temp);
   Int2Str := temp;
  end;

  procedure ReplaceStr(var S: PString; V: String);
  begin
    DisposeStr(S);
    S := NewStr(V);
  end;

begin
  FSplit(paramstr(0),ShellPath,_N,_E);
  Dec(ShellPath[0]);

  OvrInit(ShellPath+'\POVSHELL.OVR');
  if OvrResult <> ovrOk then
  begin
    Writeln('Overlay manager initialization failed.');
    case OvrResult of
      ovrError: writeln('General Error -1');
      ovrNotFound: writeln('Can''t find overlay file POVSHELL.OVR');
      ovrNoMemory: writeln('Out of memory.');
      ovrIOError: writeln('I/O-error with overlay file POVSHELL.OVR');
    end;
    Halt(1);
  end;

  GetDir(0,StartDirectory);
  TempPath := GetEnv('TEMP');
  if TempPath = '' then TempPath := ShellPath;
  if TempPath<>'' then
    if TempPath[Length(TempPath)]<>'\' then TempPath:=TempPath+'\';
  CurDEFFileName := '';
  CurPOVFileDir := '';
  UseMono := false;
  CFGPalette := '';
  Remembered := 0;
  RememberList[0].KR:=0;
  RememberList[0].Sel:=1;
  Applic[MaxApplic+1] := nil;
  for i:=MaxApplic downto 1 do begin
    Applic[i] := NewItem('(not defined)', 'Alt-'+Int2Str(i), kbAlt1+(i-1)*256,
                 cmApplic+i-1, hcUApplic,Applic[i+1]);   {+++}
    ApplicCommand[i].Prog := NewStr(' ');
    ApplicCommand[i].Params := NewStr(' ');
  end;
end.

