;**********************************************************************;
;*                            V I O S P A                             *;
;*--------------------------------------------------------------------*;
;*    Aufgabe        : Stellt eine Funktion zur Einbindung in ein     *;
;*                     Turbo-Pascal-Programm zur Verfgung, mit der   *;
;*                     die Art der beiden installierten Video-        *;
;*                     Adapter festgestellt werden kann.              *;
;*--------------------------------------------------------------------*;
;*    Autor          : MICHAEL TISCHER                                *;
;*    entwickelt am  :  2.10.1988                                     *;
;*    letztes Update : 14.01.1991                                     *;
;*--------------------------------------------------------------------*;
;*    assemblieren   : MASM VIOSPA;                                   *;
;*                     ... ber die Anweisung {$L VIOSPA} in ein      *;
;*                         TURBO-PASCAL-Programm einbinden            *;
;**********************************************************************;
                                                                        
;== Konstanten fr die Struktur VIOS ===================================
                                                                        
                                  ;Konstanten fr die Video-Karte
NO_VIOS    = 0                    ;keine Video-Karte
VGA        = 1                    ;VGA-Karte
EGA        = 2                    ;EGA-Karte
MDA        = 3                    ;Monochrome Display Adapter
HGC        = 4                    ;Hercules Graphics Card
CGA        = 5                    ;Color Graphics Adapter
                                                                        
                                  ;Konstanten fr den Monitor
NO_MON     = 0                    ;kein Monitor
MONO       = 1                    ;monochromer Monitor
COLOR      = 2                    ;Color-Monitor
EGA_HIRES  = 3                    ;hochauflsender oder Multisync-Mon.
ANAL_MONO  = 4                    ;monochromer Analog-Monitor
ANAL_COLOR = 5                    ;analoger Color-Monitor
                                                                        
;== Datensegment =======================================================
                                                                        
DATA   segment word public        ;Turbo-Datensegment
                                                                        
DATA   ends
                                                                        
;== Codesegment ========================================================
                                                                        
CODE       segment byte public    ;das Turbo-Codesegment
                                                                        
           assume cs:CODE, ds:DATA
                                                                        
public     getvios
                                                                        
;-- initiliasierte globale Variablen mssen im Code-Segment abgelegt ---
;-- werden.                                                          ---
                                                                        
vios_tab   equ this word
                                                                        
           ;-- Umsetzungstabelle fr die Rckgabewerte der Funktion ----
           ;-- 1Ah, Unterfunktion 00h des VGA-BIOS                  ----
                                                                        
           db NO_VIOS, NO_MON     ;keine Video-Karte
           db MDA    , MONO       ;MDA-Karte an monochromen Monitor
           db CGA    , COLOR      ;CGA-Karte an Color-Monitor
           db ?      , ?          ;Code 3 wird nicht benutzt
           db EGA    , EGA_HIRES  ;EGA-Karte an hochaufl. Monitor
           db EGA    , MONO       ;EGA-Karte an monochromen Monitor
           db ?      , ?          ;Code 6 wird nicht benutzt
           db VGA    , ANAL_MONO  ;VGA-Karte an analogem Mono-Monitor
           db VGA    , ANAL_COLOR ;VGA-Karte an analogem Color-Monitor
                                                                        
ega_dips   equ this byte
                                                                        
           ;-- Umsetzungstabelle fr die Dipschalter der EGA-Karte -----
                                                                        
           db COLOR, EGA_HIRES, MONO
           db COLOR, EGA_HIRES, MONO
                                                                        
;-----------------------------------------------------------------------
;-- GETVIOS: ermittelt die Art der installierten Video-Karten ---------
;-- Aufruf von C: void get_vios( struct vios *vp );
;-- Deklaration : struct vios { BYTE karte, monitor; };
;-- Return-Wert : keiner
;-- Info        : In diesem Beispiel ist die Funktion fr die Einbindung
;--               in das SMALL-Modell konzipiert
                                                                        
getvios  proc near
                                                                        
sframe     struc                  ;Struktur zum Zugriff auf den Stack
cga_possi  db ?                   ;lokale Variable
ega_possi  db ?                   ;       "
mono_possi db ?                   ;       "
bptr       dw ?                   ;nimmt BP auf
ret_adr    dw ?                   ;Rcksprungadresse zum Aufrufer
vp         dd ?                   ;Pointer auf die erste VIOS-Struktur
sframe     ends                   ;Ende der Struktur
                                                                        
frame      equ [ bp - cga_possi ] ;adressiert die Elemente der Struktur
                                                                        
           push bp                ;BP auf dem Stack sichern
           sub  sp,3              ;Platz fr lokale Variablen schaffen
           mov  bp,sp             ;SP nach BP bertragen
                                                                        
           mov  frame.cga_possi,1 ;kann CGA sein
           mov  frame.ega_possi,1 ;kann EGA sein
           mov  frame.mono_possi,1;kann MDA oder HGC sein
                                                                        
           mov  di,word ptr frame.vp     ;Offsetadr. der Struktur holen
           mov  word ptr [di],NO_VIOS    ;noch kein Video-System
           mov  word ptr [di+2],NO_VIOS  ;gefunden
                                                                        
           call test_vga          ;auf VGA-Karte testen
           cmp  frame.ega_possi,0 ;EGA-Karte noch mglich?
           je  gv1                ;NEIN --> zum CGA-Test
                                                                        
           call test_ega          ;auf EGA-Karte testen
gv1:       cmp  frame.cga_possi,0 ;CGA-Karte noch mglich?
           je   gv2               ;NEIN --> zum MDA- und HGC-Text
                                                                        
           call test_cga          ;auf CGA-Karte testen
gv2:       cmp  frame.mono_possi,0;MDA oder HGC-Karte noch mglich?
           je   gv3               ;NEIN --> Tests sind abgeschlossen
                                                                        
           call test_mono         ;auf MDA- und HGC-Karte testen
                                                                        
           ;-- aktive Video-Karte ermitteln ----------------------------
                                                                        
gv3:       cmp  byte ptr [di],VGA ;wurde ein VGA-Karte entdeckt?
           je   gvi_end           ;JA, aktive Karte bereits ermittelt
           cmp  byte ptr [di+2],VGA ;VGA-Karte als sekundres System?
           je   gvi_end           ;JA, aktive Karte bereits ermittelt
                                                                        
           mov  ah,0Fh            ;aktuellen Video-Modus ber den BIOS-
           int  10h               ;Video-Interrupt ermitteln
                                                                        
           and  al,7              ;nur die Modi 0 bis 7 sind interessant
           cmp  al,7              ;ist monochrome Karte aktiv?
           jne  gv4               ;Nein, im CGA- oder EGA-Modus
                                                                        
           ;-- die MDA, HGC-, oder EGA-Karte (mono) ist aktiv ----------
                                                                        
           cmp  byte ptr [di+1],MONO ;Mono-Monitor in erster Struktur?
           je   gvi_end           ;JA, Reihenfolge ist o.k.
           jmp  short tauschen    ;NEIN, Reihenfolge vertauschen
                                                                        
           ;-- eine CGA- oder EGA-Karte ist aktiv ----------------------
                                                                        
gv4:       cmp  byte ptr [di+1],MONO ;Mono-Monitor in erster Struktur?
           jne  gvi_end           ;NEIN, Reihenfolge ist o.k.
                                                                        
tauschen:  mov  ax,[di]           ;Inhalt der ersten Struktur holen
           xchg ax,[di+2]         ;mit zweiter Struktur tauschen
           mov  [di],ax
                                                                        
gvi_end:   add  sp,3              ;lokale Variablen vom Stack entfernen
           pop  bp                ;BP wieder vom Stack holen
           ret  4                 ;zurck zu TURBO und Variablen auf
                                  ;Stack lschen
getvios    endp
                                                                        
;-----------------------------------------------------------------------
;-- TEST_VGA: stellt fest, ob eine VGA-Karte installiert ist
                                                                        
test_vga   proc near
                                                                        
           mov  ax,1a00h          ;Funktion 1Ah, Unterfunktion 00h
           int  10h               ;des VGA-BIOS aufrufen
           cmp  al,1ah            ;wird die Funktion untersttzt?
           jne  tvga_end          ;NEIN --> Routine beenden
                                                                        
           ;-- die Funktion wurde untersttzt, BL enthlt jetzt den ----
           ;-- Code des aktiven, BH des inaktiven Video-Systems     ----
                                                                        
           mov  cx,bx             ;Resultat in CX merken
           xor  bh,bh             ;BH auf 0 setzen
           or   ch,ch             ;nur ein Video-System?
           je   tvga_1            ;JA --> Code des ersten System ausw.
                                                                        
           ;-- Code des zweiten Systems umsetzen -----------------------
                                                                        
           mov  bl,ch             ;Code des zweiten Systems nach BL
           add  bl,bl             ;Offset in Tabelle bilden
           mov  ax,vios_tab[bx]   ;Code aus Tabelle holen
           mov  [di+2],ax         ;und in Struktur des Aufrufers ablegen
           mov  bl,cl             ;Codes der ersten Sys. wieder nach BL
                                                                        
           ;-- Code des ersten Systems umsetzen ------------------------
                                                                        
tvga_1:    add  bl,bl             ;Offset in Tabelle bilden
           mov  ax,vios_tab[bx]   ;Code aus Tabelle holen
           mov  [di],ax           ;und in Struktur des Aufrufers ablegen
                                                                        
           mov  frame.cga_possi,0 ;CGA-Test nicht mehr erforderlich
           mov  frame.ega_possi,0 ;auch EGA-Test nicht
           mov  frame.mono_possi,0 ;MONO wird gleich noch geprft
                                                                        
           mov  bx,di             ;Adresse der Aktiv-Struktur
           cmp  byte ptr [bx],MDA ;monochromes System entdeckt?
           je   do_tmono          ;JA --> MDA-/HGC-Text durchfhren
                                                                        
           add  bx,2              ;Adresse der Inaktiv-Struktur
           cmp  byte ptr [bx],MDA ;monochromes System entdeckt?
           jne  tvga_end          ;NEIN Routine beenden
                                                                        
do_tmono:  mov  word ptr [bx],0   ;simulieren, da dieses System
                                  ;noch nicht entdeckt wurde
           mov  frame.mono_possi,1;monochromer Test wird durchgefhrt
                                                                        
tvga_end:  ret                    ;zurck zum Aufrufer
                                                                        
test_vga   endp
                                                                        
;-----------------------------------------------------------------------
;-- TEST_EGA: stellt fest, ob eine EGA-Karte installiert ist
                                                                        
test_ega   proc near
                                                                        
           mov  ah,12h            ;Funktion 12h
           mov  bl,10h            ;Unterfunktion 10h
           int  10h               ;des EGA-BIOS aufrufen
           cmp  bl,10h            ;wird die Funktion untersttzt?
           je   tega_end          ;NEIN --> Routine beenden
                                                                        
           ;-- die Funktion wurde untersttzt, CL enthlt jetzt die ----
           ;-- Stellung der DIP-Schalter der EGA-Karte              ----
                                                                        
           mov  bl,cl             ;DIP-Schalter nach BL
           shr  bl,1              ;eine Position nach rechts shiften
           xor  bh,bh             ;Hi-Byte fr Index auf 0
           mov  ah,ega_dips[bx]   ;Element aus Tabelle holen
           mov  al,EGA            ;ist EGA-Karte
           call gefunden          ;Daten in den Vektor eintragen
                                                                        
           cmp  ah,MONO           ;Anschlu an monochromen Bildschirm?
           je   is_mono           ;JA --> kein MDA oder HGC
                                                                        
           mov  frame.cga_possi,0 ;kein CGA-Karte mglich
           jmp  short tega_end    ;Routine beenden
                                                                        
is_mono:   mov  frame.mono_possi,0;da die EGA-Karte am monochromen Mo-
                                  ;nitor hngt, kann weder MDA noch HGC
                                  ;installiert sein
                                                                        
tega_end:  ret                    ;zurck zum Aufrufer
                                                                        
test_ega   endp
                                                                        
;-----------------------------------------------------------------------
;-- TEST_CGA: stellt fest, ob eine CGA-Karte installiert ist
                                                                        
test_cga   proc near
                                                                        
           mov  dx,3D4h           ;Portadr. des CRTC-Adressreg. bei CGA
           call test_6845         ;testen, ob 6845 CRTC installiert ist
           jc   tega_end          ;NEIN --> Test beenden
                                                                        
           mov  al,CGA            ;JA, CGA ist installiert
           mov  ah,COLOR          ;an CGA hngt immer Color-Monitor
           jmp  gefunden          ;Daten in Vektor eintragen
                                                                        
test_cga   endp
                                                                        
;-----------------------------------------------------------------------
;-- TEST_MONO: untersucht die Existenz einer MDA- oder HGC-Karte
                                                                        
test_mono  proc near
                                                                        
           mov  dx,3B4h           ;Portadr. des CRTC-Adressreg. bei MONO
           call test_6845         ;testen, ob 6845 CRTC installiert ist
           jc   tega_end          ;NEIN --> Test beenden
                                                                        
           ;-- es ist eine monochrome Bildschirmkarte installiert ------
           ;-- der folgende Test unterscheidet zwischen MDA und HGC ----
                                                                        
           mov  dl,0BAh           ;MONO-Status-Port bei 3BAh
           in   al,dx             ;Status-Port abfragen
           and  al,80h            ;nur Bit 7 stehenlassen
           mov  ah,al             ;und in AH merken
                                                                        
           ;-- wenn sich der Inhalt des Bit 7 im Statusport whrend ----
           ;-- einer der folgenden Abfragen ndert, handelt es sich ----
           ;-- um eine HGC                                          ----
                                                                        
           mov  cx,8000h          ;maximal 32768 Schleifendurchlufe
test_hgc:  in   al,dx             ;Status-Port abfragen
           and  al,80h            ;nur Bit 7 stehenlassen
           cmp  al,ah             ;hat sich der Inhalt verndert?
           jne  is_hgc            ;Bit 7 = 1 --> ist ein HGC
           loop test_hgc          ;weiterschleifen
                                                                        
           mov  al,MDA            ;Bit 7 wurde nicht 1 --> ist MDA
           jmp  set_mono          ;Parameter setzen
                                                                        
is_hgc:    mov  al,HGC            ;Bit 7 wurde 1 --> ist HGC
set_mono:  mov  ah,MONO           ;MDA und HGC grunds. an MONO-Bilds.
           jmp  gefunden          ;Parameter setzen
                                                                        
test_mono  endp
                                                                        
;-----------------------------------------------------------------------
;-- TEST_6845: kehrt mit gesetztem Carray-Flag zurck, wenn sich an der
;--            Portadresse in DX kein 6845 befindet
                                                                        
test_6845  proc near
                                                                        
           mov  al,0Ah            ;das Register 10 ansprechen
           out  dx,al             ;Registernummer an CRTC-Adressreg.
           inc  dx                ;DX jetzt auf CRTC-Datenregister
                                                                        
           in   al,dx             ;Inhalt des Register 10 holen
           mov  ah,al             ;und in AH merken
                                                                        
           mov  al,4Fh            ;beliebigen Wert
           out  dx,al             ;in das Register 10 schreiben
                                                                        
           mov  cx,100            ;kleine Warteschleife, damit der
waitv:     loop waitv             ;6845 reagieren kann
                                                                        
           in   al,dx             ;Inhalt des Register 10 wieder lesen
           xchg al,ah             ;AH und AL vertauschen
           out  dx,al             ;alten Wert zurcksenden
                                                                        
           cmp  ah,4Fh            ;wurde der geschriebene Wert gelesen?
           je   t6845_end         ;JA --> Test beenden
                                                                        
           stc                    ;NEIN --> Carry-Flag setzen
                                                                        
t6845_end: ret                    ;zurck zum Aufrufer
                                                                        
test_6845  endp
                                                                        
;-----------------------------------------------------------------------
;-- GEFUNDEN: trgt die Art der Video-Karte in AL und den Typ von  -----
;--           Monitor in AH in den Video-Vektor ein                -----
                                                                        
gefunden   proc near
                                                                        
           mov bx,di              ;Adresse der Aktiv-Struktur
           cmp word ptr [bx],0    ;bereits ein Video-System entdeckt?
           je  set_data           ;NEIN --> Daten in Aktiv-Struktur
                                                                        
           add bx,2               ;JA, Adresse der Inaktiv-Struktur
                                                                        
set_data:  mov [bx],ax            ;Daten in der Struktur ablegen
           ret                    ;zurck zum Aufrufer
                                                                        
gefunden   endp
                                                                        
;-----------------------------------------------------------------------
                                                                        
code       ends                   ;Ende des Codesegments
           end                    ;Ende des Programms
